#!/usr/bin/python3

import sys
import socket
import ssl
import traceback
from threading import Thread
from threading import Lock
from time import sleep

class ClunetPacket():
    PRIO_NOTICE=1
    PRIO_INFO=2
    PRIO_MESSAGE=3
    PRIO_COMMAND=4

    ADDR_BROADCAST=255

    CMD_DISCOVERY=0
    CMD_DISCOVERY_RESP=1
    CMD_BOOT_CONTROL=2
    CMD_REBOOT=3
    CMD_BOOT_COMPLETED=4
    CMD_PING=254
    CMD_PING_REPLY=255

    def __init__(self, prio=0, src=0, dst=None, cmd=None, data=None):
        self.prio = prio
        self.src = src
        self.dst = dst
        if self.dst == None:
            self.dst = ClunetPacket.ADDR_BROADCAST
        self.cmd = cmd
        if data == None:
            self.data = bytes()
        elif type(data) is list:
            self.data = bytes(data)
        elif type(data) is int:
            self.data = bytes([data])
        else:
            self.data = data

    @staticmethod
    def parse(line):
        line = "0" + line.strip()
        #print("Parsing '" + line + "'")
        data = bytes.fromhex(line)
        prio = data[0]
        src = data[1]
        dst = data[2]
        cmd = data[3]
        data = data[4:]
        packet = ClunetPacket(prio=prio, src=src, dst=dst, cmd=cmd, data=data)
        return packet

    def encode(self):
        return "{0:X} {1:02X} {2:02X} {3:02X} {4}".format(
            self.prio, self.src, self.dst, self.cmd, self.data.hex().upper())

    def __str__(self):
        return "prio={0}, src={1}, dst={2}, cmd={3}, data={4}".format(
            self.prio, self.src, self.dst, self.cmd, self.data.hex())


class ClunetClient():
    def __init__(self, login, password='', host='localhost', port=10010, reconnect_interval=5, ssl_context = None,
        on_data_received=None, on_connect=None, on_disconnect=None, on_error=None):
        """
        Конструктор. Сразу после создания объекта происходит подключение к серверу.
        Параметры:
            host: адрес сервера
            port: порт сервера
            password: пароль
            on_data_received: функция, которую нужно вызывать при получении данных
            on_connect: функция, которую нужно вызывать при подключении
            on_disconnect: функция, которую нужно вызывать при отключении
        """
        self._host = host
        self._port = port
        self._reconnect_interval = reconnect_interval
        self._ssl_context = ssl_context
        self._login = login
        self._password = password
        self.on_data_received = on_data_received
        self.on_connect = on_connect
        self.on_disconnect = on_disconnect
        self.on_error = on_error
        self._sock = None
        self._file_in = None
        self._working = True
        self._connection_lock = Lock()
        self.connected = False
        if self._reconnect_interval > 0:
            self._client_thread = Thread(
                target=self.connect,
                name="CLUNET connect thread",
                daemon=True)
            self._client_thread.start()

    def _reconnect(self):
        sleep(self._reconnect_interval)
        try:
            self.connect()
        except:
            pass

    def connect(self):
        with self._connection_lock:
            if self._sock: return
            try:
                self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                if self._ssl_context:
                    self._sock = self._ssl_context.wrap_socket(self._sock, server_side=False, server_hostname=self._host)
                self._sock.connect((self._host, self._port))
                self._file_in = self._sock.makefile(mode='r', encoding='utf8')
                while self._sock:
                    line = self._file_in.readline().strip(" \r\n")
                    if len(line) == 0:
                        raise ConnectionResetError()
                    if line == "LOGIN":
                        self._writeline(self._login)
                    if line == "PASSWORD":
                        self._writeline(self._password)
                    if line == "AUTH FAILED":
                        raise ClunetAuthError("access denied")
                    if line == "AUTH OK":
                        self.connected = True
                        if self.on_connect:
                            self.on_connect(self)
                        self._client_thread = Thread(
                            target=self._recv_loop,
                            name="CLUNET client loop",
                            daemon=True)
                        self._client_thread.start()
                        break
            except Exception as e:
                self._cleanup()
                if self.on_error:
                    self.on_error(self, e)
                if self._reconnect_interval > 0:
                    self._client_thread = Thread(
                        target=self._reconnect,
                        name="CLUNET reconnect thread",
                        daemon=True)
                    self._client_thread.start()
                raise e

    def send(self, packet):
        if not self.connected:
            raise ClunetNotConnectedError("CLUNET is not connected")
        self._writeline(packet.encode())

    def _writeline(self, line):
        self._sock.send((line + "\r\n").encode("utf8"))

    def _recv_loop(self):
        """
        Внутренний цикл для чтения данных, выполняется фоном в отдельном потоке.
        """
        try:
            while self._sock:
                line = self._file_in.readline().strip(" \r\n")
                if len(line) == 0:
                    break
                line = line.upper()
                packet = ClunetPacket.parse(line)
                if self.on_data_received:
                    self.on_data_received(self, packet)
        except Exception as e:
            if self.on_error:
                self.on_error(self, e)
        finally:
            self._cleanup()
            if self.connected and self.on_disconnect:
                self.on_disconnect(self)
            if self._reconnect_interval > 0:
                self._client_thread = Thread(
                    target=self._reconnect,
                    name="CLUNET reconnect thread",
                    daemon=True)
                self._client_thread.start()

    def _cleanup(self):
        if self._sock:
            try:
                self._sock.shutdown(socket.SHUT_RDWR)
            except:
                pass
            try:
                self._sock.close()
            except:
                pass
        if self._file_in:
            try:
                self._file_in.close()
            except:
                pass
        self._sock = None
        self._file_in = None
        self.connected = False

    def close(self):
        """
        Разрывает соединение, завершает работу
        """
        self._working = False
        self._cleanup()

    def __del__(self):
       self.close()

class ClunetNotConnectedError(Exception):
    pass


class ClunetAuthError(Exception):
    pass
