#!/usr/bin/python3

import sys
import users
import socket
import ssl
from ipaddress import ip_network
from threading import Thread
from threading import Lock
from time import sleep
from hashlib import md5
from importlib import reload
from clunet_client import ClunetPacket
from clunet_client import ClunetClient
from clunet_serial import ClunetSerial
from clunet_gpio_driver import ClunetGpioDriver
import traceback

class ClunetServer:
    def __init__(self, clunet_device, port = 10010, interface = '0.0.0.0', self_address = 0, port_ssl = None, ssl_context = None):
        self._clunet_device = clunet_device
        self._port = port
        self._interface = interface
        self._self_address = self_address
        self._port_ssl = port_ssl
        self._ssl_context = ssl_context
        self._working = False
        self._clients = []
        self._server_sock = None
        self._server_sock_ssl = None

    def start(self):
        self._working = True
        self._server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server_sock.bind((self._interface, self._port))
        self._server_sock.listen(10)
        print("Listening port", self._port)
        Thread(target=self._port_listener, name="CLUNET port listener", daemon=True).start()

        if self._port_ssl:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._server_sock_ssl = self._ssl_context.wrap_socket(sock, server_side=True)
            self._server_sock_ssl.bind((self._interface, self._port_ssl))
            self._server_sock_ssl.listen(10)
            print("Listening SSL port", self._port_ssl)
            Thread(target=self._port_listener_ssl, name="CLUNET SSL port listener", daemon=True).start()

        Thread(target=self._device_listener, name="CLUNET device listener", daemon=True).start()

    def stop(self):
        self._working = False
        if self._server_sock != None:
            try:
                self._server_sock.close()
            except:
                pass
            self._server_sock = None
        if self._server_sock_ssl != None:
            try:
                self._server_sock_ssl.close()
            except:
                pass
            self._server_sock_ssl = None
        for client in self._clients:
            client.close()

    def __del__(self):
        self.stop()

    def _port_listener(self):
        while self._working and self._server_sock:
            try:
                conn, addr = self._server_sock.accept()
                print('Connected: {0}'.format(addr))
                ClunetClientConnection(self, conn, addr)
            except Exception as e:
                print("Error:", e)

    def _port_listener_ssl(self):
        while self._working and self._server_sock_ssl:
            try:
                conn, addr = self._server_sock_ssl.accept()
                print('Connected (using SSL): {0}'.format(addr))
                ClunetClientConnection(self, conn, addr)
            except Exception as e:
                print("Error:", e)

    @staticmethod
    def check_crc(data):
        crc = 0
        for i in range(len(data)):
            inbyte = data[i]
            for j in range(8):
                mix = (crc ^ inbyte) & 1
                crc = crc >> 1
                if mix != 0:
                    crc = crc ^ 0x8C
                inbyte = inbyte >> 1
        return crc

    def _device_listener(self):
        while self._working:
            try:                
                packet = self._clunet_device.recv()
                print("Received from bus:", packet)
                for client in self._clients:
                    if (packet.src in client.devices_can_read) and (
                        client.allow_sniff or packet.dst == self._self_address or
                        packet.dst == ClunetPacket.ADDR_BROADCAST) and (
                        packet.cmd in client.commands_can_read):
                            try:
                                client.writeline(packet.encode())
                            except Exception as e:
                                print("Error:", e)
            except Exception as e:
                print("Error:", e)

    def write_to_bus(self, packet):
        self._clunet_device.send(packet)

class ClunetClientConnection:
    def __init__(self, server, socket, address):
        self.server = server
        self._sock = socket
        self.address = address
        self.auth = False
        self.devices_can_read = set()
        self.devices_can_write = set()
        self.commands_can_read = set()
        self.commands_can_write = set()
        self.allow_sniff = False
        self.file_in = socket.makefile(mode='r', encoding='utf8')
        self.file_out = socket.makefile(mode='w', encoding='utf8')
        self.c_thread = Thread(target = self.client_thread, daemon = True)
        self.c_thread.start()

    def _readline(self):
        line = self.file_in.readline().strip("\r\n")
        return line

    def writeline(self, line):
        print(line, file=self.file_out, flush=True)

    def client_thread(self):
        try:
            self.writeline("LOGIN")
            login = self._readline()
            self.writeline("PASSWORD")
            password = md5(self._readline().encode("utf-8")).hexdigest()
            ip = ip_network(self.address)
            reload(users) # Reload users database
            for user in users.users:
                if user["login"] == login and user["password"] == password and ip.overlaps(user["network"]):
                    self.auth = True
                    if "devices_can_read" in user:
                        self.devices_can_read = set(list(self.devices_can_read) + list(user["devices_can_read"]))
                    if "devices_can_write" in user:
                        self.devices_can_write = set(list(self.devices_can_write) + list(user["devices_can_write"]))
                    if "commands_can_read" in user:
                        self.commands_can_read = set(list(self.commands_can_read) + list(user["commands_can_read"]))
                    if "commands_can_write" in user:
                        self.commands_can_write = set(list(self.commands_can_write) + list(user["commands_can_write"]))
                    if "allow_sniff" in user and user["allow_sniff"]:
                        self.allow_sniff = True

            if self.auth:
                self.writeline("AUTH OK")
                self.server._clients.append(self)
            else:
                print("Auth failed for user {0}".format(login))
                sleep(3)
                self.writeline("AUTH FAILED")
                return

            while self._sock and self.server._working:
                line = self._readline()
                if len(line) == 0:
                    break
                packet = ClunetPacket.parse(line)
                print("Received from {0}: {1}".format(self.address, packet))
                if packet.dst not in self.devices_can_write and ClunetPacket.ADDR_BROADCAST not in self.devices_can_write: 
                    print("User \"{0}\" can't write to {1} (0x{1:02X})".format(login, packet.dst))
                    continue
                if packet.cmd not in self.commands_can_write:
                    print("User \"{0}\" can't write {1} (0x{1:02X}) command".format(login, packet.cmd))
                    continue
                self.server.write_to_bus(packet)

        except Exception as e:
            print("Error:", e)
            traceback.print_exc()
        finally:
            print("Disconnected: {0}".format(self.address))
            if self in self.server._clients:
                self.server._clients.remove(self)
            if self._sock != None:
                try:
                    self._sock.shutdown(socket.SHUT_RDWR)
                    self._sock.close()
                except:
                    pass
            if self.file_out != None:
                try:
                    self.file_out.close()
                except:
                    pass
            if self.file_in != None:
                try:
                    self.file_in.close()
                except:
                    pass
            self._sock = None
            self.file_out = None
            self.file_in = None

    def close(self):
        try:
            self._sock.close()
            self._sock = None
        except:
            pass

    def __del__(self):
        self.close()

if __name__ == "__main__":
    clunet_device = ClunetSerial("/dev/clunet", 115200)
    #clunet_device = ClunetGpioDriver("/dev/clunet")

    CLUNET_INTERFACE = 10010
    CLUNET_PORT = 10010
    CLUNET_PORT_SSL = 10011
    CLUNET_SELF_ADDRESS = 0
    FULLCHAIN_KEY = "/home/clunet/.keys/fullchain.pem"
    PRIVATE_KEY = "/home/clunet/.keys/privkey.pem"

    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_OPTIONAL
    ssl_context.load_cert_chain(FULLCHAIN_KEY, PRIVATE_KEY)

    server = ClunetServer(clunet_device, port=CLUNET_PORT, port_ssl=CLUNET_PORT_SSL, ssl_context=ssl_context)
    server.start()
    while True:
        sleep(3600)
        ssl_context.load_cert_chain(FULLCHAIN_KEY, PRIVATE_KEY)
